"""
    Definition of the backward-compatible CLI interface of PyPhoneBook.
"""
from typing import Any, Dict, List, Tuple
import sys
import traceback
from textwrap import dedent
from pyphonebook.cli_parser import getParser
from pyphonebook import PhoneBook, PhoneBookEntry, PHONEBOOK_AND, PHONEBOOK_OR, _TERSE_ALIASES


_ORDERBY_ALIASES = {"surname": "last_name",
                   "firstname": "first_name",
                   "login": "login_name",
                   "department": "department",
                   "group": "_extensionAttribute13",
                   "cerngroup": "group",
                   "cernsection": "section",
                   "displayname": "_displayName"}

def _cli() -> None:
    """
    Defines the CLI interface with ``argparse`` and uses
    ``PhoneBook.compound_search()`` to print the results.

    Output formatting is not done here: see PhoneBookEntry.__str__()

    Can be used a simple usage example.
    """
    verbose = True  # In case it fails before reading the flag
    try:
        parser = getParser()

        # ######################################## #
        #                 PARSING                  #
        # ######################################## #

        # Gather the parameters from the command line as a dictionary
        parameters = vars(parser.parse_args())

        # Print warning for ambiguous flags - they are going to be ignored
        for flag, ambiguity in [
            ('o', "office, orderby"),
            ('e', "email, external"),
            ('s', "surname, section"),
            ('p', "phone, pobox"),
            ('g', "gid, group"),
            ('h', "help, homedir")
        ]:
            if parameters.pop(flag, False):
                print(f"Option {flag} is ambiguous ({ambiguity})", file=sys.stderr)

        # Filter out non-search related flags for later use
        search_string = parameters.pop('search_string', None)
        if search_string:
            search_string = ' '.join(search_string)
        externals = parameters.pop('external', None)
        yellowpages = parameters.pop('yellowpages', None)
        all_fields = parameters.pop('all', None)
        orderby = parameters.pop('orderby', [])
        terse = parameters.pop('terse', [])
        json = parameters.pop('json', [])
        verbose = parameters.pop('verbose', False)

        # Identify the search domain
        domain = "users"
        if externals:
            domain = "externals"
        elif yellowpages:
            domain = "services"

        # Validate terse fields, if needed:
        for field in terse:
            if field not in _TERSE_ALIASES.keys():
                print(f'Error: Unrecognized -t(erse) option value ({field}), see --help for help.', file=sys.stderr)
                return 2

        # Validate json fields, if needed:
        for field in json:
            if field not in _TERSE_ALIASES.keys():
                print(f'Error: Unrecognized -j(son) option value ({field}), see --help for help.', file=sys.stderr)
                return 2

        # Validate orderby fields, if needed:
        for field in orderby:
            if field not in _ORDERBY_ALIASES.keys():
                print(f'Error: Unrecognized -order(by) option value ({field}), see --help for help.', file=sys.stderr)
                return 2

        # Instantiate a PhoneBook. Note: it can throw LDAPErrors in case of failure (i.e. network failures)
        pb = PhoneBook(domain=domain)

        # If terse or json is set, we don't want to merge entries by CCID anymore (that's the behavior of the old CLI)
        merge = (len(terse) == 0 and len(json) == 0)

        # If we have nothing to do, just show the help
        if not parameters and not search_string:
            parser.print_help()
            return 1

        # Build the query and search
        query_parameters = _build_query(parameters, search_string)
        matches = pb.arbitrary_search(parameters=query_parameters, raw=False, merge_by_ccid=merge, verbose=verbose)

        # If no results were found, return status code 1
        if len(matches) == 0:
            if json:
                print('[]')
            return 1

        # Sort the matches, first by login so multiple iterations return results in consistant order
        _sort_by_fields(matches, ['login'])
        if orderby:
            _sort_by_fields(matches, orderby)
        else:
            _sort_by_fields(matches, ['surname', 'firstname'])

        # Print the entries, properly rendered
        output = [match._render(terse=terse, json=json, all_fields=all_fields) for match in matches]
        if json:
            print(f'[{",".join(output)}]')
        else:
            [print(x) for x in output]
        if all_fields and not yellowpages:
            print(dedent("""\
                            #-------------------------------------------------------------------------------
                            #Account St(atus): P(rimary), S(econdary), s(ervice), U(nknown)
                            #                  A(ctive),  D(isabled),  P(assword expired),  L(ocked out)"""))
        elif terse:
            print("#-------------------------------------------------------------------------------\n"
                  "WARNING! If you built a script on top of the PhoneBook CLI, please consider\n"
                  "using the PhoneBook Python API instead.\n"
                  "#-------------------------------------------------------------------------------", file=sys.stderr)

    except Exception as e:
        if verbose:
            traceback.print_exc()
        print("-" * 80, file=sys.stderr)
        print("A fatal exception occurred!", file=sys.stderr)
        print(e, file=sys.stderr)
        print("-" * 80, file=sys.stderr)
        return 30

    # Everything went well!
    return 0


def _build_query(parameters: Dict[str, Any], search_string: str = None) -> Tuple[str, Any]:
    """
    Given the search_string and all the other explicit flags, build the correct query for the search.

    :param parameters: all the explicit fields given in the command line, like ``--firstname John``
    :param search_string: the standalone string given without specifying any flag, usually a login name or phone
        number.
    :return: a well-formed ``PhoneBook`` query.
    """
    query_parameters_list = []

    # A free search string is interpreted as follows:
    #    - If it's numeric, look for a phone number (either long or short form)
    #    - If it isn't, look for a display name or account name
    if search_string:
        if search_string.strip().strip("+").isdigit():  # Phone numbers might begin with +, so check that too
            search_string_parameters = (PHONEBOOK_OR, [
                ("mobile", f"*{PhoneBookEntry.short_phone_number_to_long(search_string)}*"),
                ("mobile", f"*{PhoneBookEntry.long_phone_number_to_short(search_string)}*"),
                ("telephoneNumber", f"*{PhoneBookEntry.short_phone_number_to_long(search_string)}*"),
                ("telephoneNumber", f"*{PhoneBookEntry.long_phone_number_to_short(search_string)}*")
            ])
        else:
            search_string_parameters = (PHONEBOOK_OR, [
                ("displayName", f"*{search_string}*"),
                ("sAMAccountName", search_string)
            ])
        query_parameters_list.append(search_string_parameters)

    # Append all the other parameters as a list of tuples
    query_parameters_list.extend(parameters.items())

    # AND all the parameters
    complete_query = (PHONEBOOK_AND, query_parameters_list)

    return complete_query


def _sort_by_fields(matches: List[PhoneBookEntry], fields: List[str]) -> List[PhoneBookEntry]:
    """
    Mimick the old phonebook --orderby flag behavior. Orders the list of entries by the specified fields.
    :param matches: list of PhoneBook entries
    :param fields: Fields to sort the above list by
    :return: an ordered list of PhoneBookEntry objects
    """
    # Double check: technically the fields have already been validated
    if any(field not in _ORDERBY_ALIASES.keys() for field in fields):
        raise ValueError(f"Valid fields for orderby are: {_ORDERBY_ALIASES.keys()}")

    for field in reversed(fields):
        matches.sort(key=lambda entry: __safelyGetSortKey(entry, _ORDERBY_ALIASES[field] ) )
    return matches

def __safelyGetSortKey(entry, key):
    try:
        return vars(entry)[key][0].lower()
    except (KeyError, IndexError):
        return "-"

def main():
    try:
        sys.exit(_cli())
    except KeyboardInterrupt:
        sys.exit(130)

if __name__ == "__main__":
    main()
