"""
Perform queries to the specified CERN LDAP database.
"""

from typing import Any, List, Mapping, Tuple, Union
import ldap3
from functools import lru_cache

from pyphonebook.phonebook_entry import PhoneBookEntry


#: LDAP symbols for AND
PHONEBOOK_AND = "&"
#: LDAP symbol for OR
PHONEBOOK_OR = "|"

#: Dictionary of simple, predefined queries that can be reused by this class or by the end user as well;
#: good to be fed to ``PhoneBook.search()`` or ``PhoneBook._raw_search()``
QUERIES = {
    'login_name':   "(sAMAccountName={0})",
    'full_name':    "(displayName={0})",
    'first_name':   "(givenName={0})",
    'last_name':    "(sn={0})",
    'email':        "(|(mail={0})(proxyAddresses=smtp:{0}))",
    'phone':        "(|(telephoneNumber=*{0})(mobile=*{0})(facsimileTelephoneNumber=*{0}))",
    'landline_phone': "(telephoneNumber=*{0})",
    'fax':          "(facsimileTelephoneNumber=*{0})",
    'mobile_phone': "(mobile=*{0})",
    'post_box':     "(postOfficeBox={0})",
    'building':     "(|(physicalDeliveryOfficeName=Bld*{0} *)(physicalDeliveryOfficeName={0} *))",
    'office':       "(|(physicalDeliveryOfficeName=*Room*{0}*)(physicalDeliveryOfficeName=* *{0}*))",
    'department':   "(division={0})",
    'group':        "(department=*/{0})",
    'section':      "(cernSection={0})",
    'organizational_unit': "(&(division={0})(department=*/{1})(cernSection={2}))",
    'employee_id':  "(employeeID={0})",
    'ccid':         "(employeeID={0})",
    'uid':          "(uidNumber={0})",
    'gid':          "(gidNumber={0})",
    'home_directory': "(unixHomeDirectory={0})",
}

#: LDAP server
LDAP_SERVER = 'ldap://xldap.cern.ch'

#: Known LDAP domains where PhoneBook can look for entries.
LDAP_DOMAINS = {
    "users": "ou=users,ou=organic units,dc=cern,dc=ch",
    "externals": "ou=externals,dc=cern,dc=ch",
    "services": "ou=Services,dc=cern,dc=ch"
}

#: Queries and fields known to contain a phone number
KNOWN_PHONE_FIELDS = [
    'phone',
    'landline_phone',
    'fax',
    'mobile_phone',
    'telephoneNumber',
    'facsimileTelephoneNumber',
    'mobile',
    'otherTelephone',
]


class PhoneBook:
    """
    Performs queries to the specified CERN LDAP database. Offers some shorthands for frequent queries.

    Query results are parsed by default into ``PhoneBookEntry`` objects for ease of access,
    but such feature can be disable for a faster lookup using the ``raw`` flag.

    If the ``raw`` flag is set, the results are dictionaries containing strings, byte-encoded strings,
    or LDAP references. Otherwise, every field is guaranteed to be a list of strings, the most common fields
    are renamed for convenience (while the all the original fields are kept with an underscore prefixed to their name)
    list of LDAP references are coarsely parsed, and known date fields are parsed too.
    """

    def __init__(self, domain: str = "users", server: str = LDAP_SERVER):
        """
        Initializes the LDAP connection. Might throw ``ldap.LDAPError`` if the connection fails.

        :param domain: DN to bind the LDAP folder to. Defaults to ``users`` but can be set to ``externals`` or ``services`` for special searches.
        :param server: LDAP server address, Defaults to "ldap://xldap.cern.ch".
        :return: the LDAP connection object, already bind.
        """
        self.domain, self.domain_name = None, None
        self.set_domain(domain)
        self.connection = self._getLDAP(server)

    def _getLDAP(self, server: str):
        """
        Get the LDAP connection. Separate function for easier mocking
        """
        return ldap3.Connection(server, auto_bind=True)

    def set_domain(self, domain: str):
        """
        Translates a user-friendly domain name into its LDAP counterpart and assigns it to ``self.domain``
        (the actual LDAP domain) and ``self.domain_name`` (the user-friendly name).

        :param domain: any key value of the ``DOMAINS`` dictionary.
        """
        if domain.lower() not in LDAP_DOMAINS.keys():
            raise ValueError(f"Unrecognized domain value: valid values are {LDAP_DOMAINS.keys()}")
        self.domain = LDAP_DOMAINS[domain.lower()]
        self.domain_name = domain

    def _raw_search(self, query: str, search_base: str = None, attributes: List = ldap3.ALL_ATTRIBUTES) -> Union[List[PhoneBookEntry], List[Any]]:
        """
        Performs a paginated query to the LDAP server using the given query directly.
        Might fail if the query is incorrectly formatted, or the fields used do not exist.

        Cheaper than ``self.search()``, as it does not perform the parsing of the retrieved entries.
        Useful for validation purposes.

        :param query: The field to send to the LDAP server to retrieve data.
        :return: Returns the unformatted list of entries.
        """
        results = []
        if not search_base:
            search_base = self.domain
        entry_generator = self.connection.extend.standard.paged_search(search_base,
                                                 search_filter = query,
                                                 search_scope = ldap3.SUBTREE,
                                                 attributes = attributes,
                                                 paged_size = 1000,
                                                 generator=True)
        for entry in entry_generator:
            results.append((entry['dn'], entry['attributes']))

        return results

    def search(self, query: str,
               raw: bool = False,
               verbose: bool = False,
               merge_by_ccid: bool = True) -> Union[List[PhoneBookEntry], List[Any]]:
        """
        Performs a query to the LDAP server using the given query directly.
        Might fail if the query is incorrectly formatted, or the fields used do not exist.

        :param query: The field to send to the LDAP server to retrieve data
        :param raw: Whether to parse the output results into ``PhoneBookEntries`` or not
        :param verbose: Whether to print out a stacktrace in case an error occurs
        :param merge_by_ccid: merges all the parser ``PhoneBookEntry`` in such a way that there will be no entries
            with the same CCID in the returned list. Ignored if ``raw == True``.
        :return: Returns a list of ``PhoneBookEntry`` (or the raw LDAP results if ``raw=True``)
        """
        results = self._raw_search(query)
        # Augment the search results with the group name for the given gidNumber
        for entry in results:
            if entry[1].get('gidNumber', False):
                try:
                    groupName = self.get_group_info(entry[1]['gidNumber'][0])['name']
                except TypeError:
                    groupName = None
                entry[1]['gidName'] = groupName
        if raw:
            return results
        else:
            entries = [PhoneBookEntry(entry, domain=self.domain_name, verbose=verbose) for entry in results]
            if merge_by_ccid:
                return self._merge_entries_by_ccid(entries)
            return entries

    def search_by_ldap_parameter(self, search_parameter: str, value: str, **kwargs) \
            -> Union[List[PhoneBookEntry], List[Any]]:
        """
        Performs a query to the LDAP server using one explicit field name.

        Extra **kwargs are passed to ``self.search()``.

        :param search_parameter:  The reference parameter to search matches on.
        :param value: The value of the parameter to search a match with.
        :return: Returns a list of ``PhoneBookEntry`` (or the raw LDAP results if ``raw=True``).
        """
        query = "(&({}={}))".format(search_parameter, value)
        return self.search(query, **kwargs)

    def multi_field_search(self, parameters: Mapping[str, str], **kwargs) -> Union[List[PhoneBookEntry], List[Any]]:
        """
        Given a list of (key, value) pairs, performs a search against all these parameters at once (ANDs them).

        To have a mix of ANDs and ORs, use ``arbitrary_search()``.

        Extra **kwargs are passed to ``self.search()``.

        :param parameters: dict of parameters for the search, like ``{'first_name': 'John', 'group': 'BI'}``.
        Note that a key might either:

         - be present in ``FLAG_ALIASES`` or ``QUERIES``, in which case is replaced with the matching query, or

         - be a simple LDAP field, in which case is directly put into a ``(key=value)`` string before chaining
           it with the other fields.

        This is useful to chain complex queries, like phone numbers.
        :return: list of ``PhoneBookEntry`` that matches all the parameters (or the raw LDAP results if ``raw=True``).
        """
        query_parameters = (PHONEBOOK_AND, list(parameters.items()))
        query = self.build_arbitrary_query(query_parameters)
        return self.search(query, **kwargs)

    def arbitrary_search(self, parameters: Tuple[str, Union[List, str]], **kwargs) -> Union[List[PhoneBookEntry], List[Any]]:
        """
        Given a dictionary or nested parameters, returns a query matching the specified arbitrary search.
        Note that a key might either:

        - be present in ``FLAG_ALIASES`` or ``QUERIES``, in which case is replaced with the matching query, or

        - be a simple LDAP field, in which case is directly put into a ``(key=value)`` string before chaining
          it with the other fields.

        This is useful to chain complex queries, like phone numbers.

        Extra kwargs are passed to ``self.search()``.

        :param parameters: The search parameters. They are interpreted by ``self.build_arbitrary_query()``.
        :return: list of ``PhoneBookEntry`` that matches all the parameters.
        """
        return self.search(self.build_arbitrary_query(parameters), **kwargs)

    @staticmethod
    def build_arbitrary_query(parameters: Tuple[str, Union[List, str]], interpret_phone_numbers: bool = True) -> str:
        """
        Given a dictionary or nested parameters, returns a query matching the specified arbitrary search.

        Note that a key might either:

         - be present in ``FLAG_ALIASES`` or ``QUERIES``, in which case is replaced with the matching query, or

         - be a simple LDAP field, in which case is directly put into a ``(key=value)`` string before chaining
           it with the other fields.

        This is useful to chain complex queries, like phone numbers. Note that phone numbers are searched in both
        forms (long and short) automatically. You can force this function to look for only one of the two forms
        by setting the parameter ``interpret_phone_numbers`` to ``False``

        :param parameters: The search parameters. Expected structure::

            (OR: [
                (AND: [
                    ("last_name", "smi")
                    ('first_name", "Jo")
                ]),
                ("department": "BE"),
                ("department": "SY")
            ])

        Resulting query: ``(|(&(sn=smi)(firstname=Jo))(department=BE)(department=SY))``
        Meaning: Search for entries with either firstname=Jo and sn=smi, or with department=BE. or with department=SY
        :param interpret_phone_numbers: search for phone numbers both in long and short form.
        :return: The query to be used to perform the desired search
        """
        key, value = parameters

        if isinstance(value, list):
            if key != PHONEBOOK_AND and key != PHONEBOOK_OR:
                raise ValueError(f"Only tuples whose first element is {PHONEBOOK_AND} or {PHONEBOOK_OR} "
                                 f"can be followed by a list.")
            query = ""
            for pair in value:
                query += PhoneBook.build_arbitrary_query(pair, interpret_phone_numbers=interpret_phone_numbers)
            return f"({key}{query})"

        # Phone number should be searched in short and long form
        elif key in KNOWN_PHONE_FIELDS and interpret_phone_numbers:
            parameters = (PHONEBOOK_OR, [
                            (key, PhoneBookEntry.short_phone_number_to_long(value)),
                            (key, PhoneBookEntry.long_phone_number_to_short(value))
                         ])
            return PhoneBook.build_arbitrary_query(parameters, interpret_phone_numbers=False)

        else:
            if value is not None and value != "":

                if key in QUERIES.keys():
                    # If it's a key present in QUERIES, replace the value with the corresponding query
                    return QUERIES[key].format(value)
                else:
                    # Otherwise just use the raw key name
                    return f"({key}={value})"

    @staticmethod
    def _merge_entries_by_ccid(entries: List[PhoneBookEntry]) -> List[PhoneBookEntry]:
        """
        Merge duplicate entries grouping them by CCID (as the old phonebook does).
        Note that entries with no CCID (like many services) are also returned unmodified.

        :return: a list of ``PhoneBookEntry`` where there are no entries with the same CCID.
        """
        # Filter out entries with no CCID: there might be some!
        no_ccid_entries = [entry for entry in entries if not entry.ccid]

        # Create a dict with this structure: { ccid: [entry1, entry2], ... }
        entries_by_ccid = {ccid: [] for entry in entries for ccid in entry.ccid}
        for entry in entries:
            for ccid in entry.ccid:
                entries_by_ccid[ccid].append(entry)

        # Merge all items that have the same ccid into one
        for ccid, entries in entries_by_ccid.items():

            # Take no action if there are 0 or 1 entry
            if len(entries) == 0:
                continue
            if len(entries) == 1:
                entries_by_ccid[ccid] = entries_by_ccid[ccid][0]
                continue

            # Merge if there is more than 1 entry
            # First sort the entries by their account types, aka. their relevance
            entries.sort(key=lambda e: {'Primary': 0, 'Secondary': 1, 'Service': 2}.get(e.account_type[0], 3) if e.account_type else 4)
            merged_entry = entries[0]
            for entry in entries[1:]:

                # Attributes might be lists on either side, so try to make sure that
                # the resulting list is always flat and always a list
                for attribute_name, attribute_value in vars(entry).items():
                    if not attribute_value:
                        continue

                    values = getattr(merged_entry, attribute_name)
                    if values is None:
                        values = attribute_value
                    else:
                        try:
                            values.extend(attribute_value)
                        except TypeError:
                            values.append(attribute_value)
                        except AttributeError:
                            values = [values, attribute_value]

                    # Remove potential duplicates (apart from selected fields)
                    try:
                        values = list(dict.fromkeys(values))
                    except TypeError:
                        # Somehow values is not an iterable
                        values = [values]
                    setattr(merged_entry, attribute_name, values)

                # parsing_failed is the only attribute not supposed to be a list in any case
                try:
                    entry.parsing_succeeded = any(entry.parsing_succeeded)
                except TypeError:
                    # Could happen if for some reason parsing_failed is not a list at this point
                    pass

            entries_by_ccid[ccid] = merged_entry

        return list(entries_by_ccid.values()) + no_ccid_entries

    def validate_login_name(self, login: str) -> bool:
        """
        Given a login name, makes sure it exists on the LDAP server (i.e. is valid)

        :param login: The login to validate
        :return: Returns ``True`` if the LDAP server returns any result for the given login name, ``False`` otherwise.
        """
        return len(self.search("(&{})".format(QUERIES['login_name'], raw=True).format(login))) > 0

    def search_by_login_name(self, value: str, raw: bool = False) -> Union[List[PhoneBookEntry], List[Any]]:
        """
        Returns a list of ``PhoneBookEntry`` for each entity in the PhoneBook with a matching login name.
        """
        return self.search("(&{})".format(QUERIES['login_name']).format(value), raw=raw)

    def search_by_full_name(self, value: str, raw: bool = False) -> Union[List[PhoneBookEntry], List[Any]]:
        """
        Returns a list of ``PhoneBookEntry`` for each entity in the PhoneBook with a matching full name.
        """
        return self.search("(&{})".format(QUERIES['full_name']).format(value), raw=raw)

    def search_by_first_name(self, value: str, raw: bool = False) -> Union[List[PhoneBookEntry], List[Any]]:
        """
        Returns a list of ``PhoneBookEntry`` for each entity in the PhoneBook with a matching first name.
        """
        return self.search("(&{})".format(QUERIES['first_name']).format(value), raw=raw)

    def search_by_last_name(self, value: str, raw: bool = False) -> Union[List[PhoneBookEntry], List[Any]]:
        """
        Returns a list of ``PhoneBookEntry`` for each entity in the PhoneBook with a matching last name.
        """
        return self.search("(&{})".format(QUERIES['last_name']).format(value), raw=raw)

    def search_by_email(self, value: str, raw: bool = False) -> Union[List[PhoneBookEntry], List[Any]]:
        """
        Returns a list of ``PhoneBookEntry`` for each entity in the PhoneBook with a matching email.
        """
        return self.search("(&{})".format(QUERIES['email']).format(value), raw=raw)

    def search_by_phone_number(self, value: str, raw: bool = False) -> Union[List[PhoneBookEntry], List[Any]]:
        """
        Returns a list of ``PhoneBookEntry`` for each entity in the PhoneBook with a matching phone number.
        """
        query_short = QUERIES['phone'].format(PhoneBookEntry.short_phone_number_to_long(value))
        query_long = QUERIES['phone'].format(PhoneBookEntry.long_phone_number_to_short(value))
        return self.search(f"(&(|{query_short}{query_long}))", raw=raw)

    def search_by_landline(self, value: str, raw: bool = False) -> Union[List[PhoneBookEntry], List[Any]]:
        """
        Returns a list of ``PhoneBookEntry`` for each entity in the PhoneBook with a matching landline phone number.
        """
        query_short = QUERIES['landline_phone'].format(PhoneBookEntry.short_phone_number_to_long(value))
        query_long = QUERIES['landline_phone'].format(PhoneBookEntry.long_phone_number_to_short(value))
        return self.search(f"(&(|{query_short}{query_long}))", raw=raw)

    def search_by_fax(self, value: str, raw: bool = False) -> Union[List[PhoneBookEntry], List[Any]]:
        """
        Returns a list of ``PhoneBookEntry`` for each entity in the PhoneBook with a matching fax.
        """
        query_short = QUERIES['fax'].format(PhoneBookEntry.short_phone_number_to_long(value))
        query_long = QUERIES['fax'].format(PhoneBookEntry.long_phone_number_to_short(value))
        return self.search(f"(&(|{query_short}{query_long}))", raw=raw)

    def search_by_mobile(self, value: str, raw: bool = False) -> Union[List[PhoneBookEntry], List[Any]]:
        """
        Returns a list of ``PhoneBookEntry`` for each entity in the PhoneBook with a matching mobile phone number.
        """
        query_short = QUERIES['mobile_phone'].format(PhoneBookEntry.short_phone_number_to_long(value))
        query_long = QUERIES['mobile_phone'].format(PhoneBookEntry.long_phone_number_to_short(value))
        return self.search(f"(&(|{query_short}{query_long}))", raw=raw)

    def search_by_post_box(self, value: str, raw: bool = False) -> Union[List[PhoneBookEntry], List[Any]]:
        """
        Returns a list of ``PhoneBookEntry`` for each entity in the PhoneBook with a matching post box.
        """
        return self.search("(&{})".format(QUERIES['post_box']).format(value), raw=raw)

    def search_by_building(self, value: str, raw: bool = False) -> Union[List[PhoneBookEntry], List[Any]]:
        """
        Returns a list of ``PhoneBookEntry`` for each entity in the PhoneBook with a matching building.
        """
        return self.search("(&{})".format(QUERIES['building']).format(value), raw=raw)

    def search_by_office(self, value: str, raw: bool = False) -> Union[List[PhoneBookEntry], List[Any]]:
        """
        Returns a list of ``PhoneBookEntry`` for each entity in the PhoneBook with a matching office.
        """
        return self.search("(&{})".format(QUERIES['office']).format(value), raw=raw)

    def search_by_department(self, value: str, raw: bool = False) -> Union[List[PhoneBookEntry], List[Any]]:
        """
        Returns a list of ``PhoneBookEntry`` for each entity in the PhoneBook with a matching department code (i.e. BE).
        """
        return self.search("(&{})".format(QUERIES['department']).format(value), raw=raw)

    def search_by_group(self, value: str, raw: bool = False) -> Union[List[PhoneBookEntry], List[Any]]:
        """
        Returns a list of ``PhoneBookEntry`` for each entity in the PhoneBook with a matching group code (i.e. BI).
        """
        return self.search("(&{})".format(QUERIES['group']).format(value), raw=raw)

    def search_by_section(self, value: str, raw: bool = False) -> Union[List[PhoneBookEntry], List[Any]]:
        """
        Returns a list of ``PhoneBookEntry`` for each entity in the PhoneBook with a matching section code (i.e. SW).
        """
        return self.search("(&{})".format(QUERIES['section']).format(value), raw=raw)

    def search_by_organizational_unit(self, department: str, group: str, section: str) \
            -> Union[List[PhoneBookEntry], List[Any]]:
        """
        Returns a list of ``PhoneBookEntry`` for each entity in the PhoneBook with a matching
        department, group and section codes (i.e. BE, BI, SW).
        """
        return self.search("(&{})".format(QUERIES['organizational_unit']).format(department, group, section))

    def search_by_employee_id(self, value: str, raw: bool = False) -> Union[List[PhoneBookEntry], List[Any]]:
        """
        Returns a list of ``PhoneBookEntry`` for each entity in the PhoneBook with a matching employee ID.
        """
        return self.search("(&{})".format(QUERIES['employee_id']).format(value), raw=raw)

    def search_by_ccid(self, value: str, raw: bool = False) -> Union[List[PhoneBookEntry], List[Any]]:
        """
        Returns a list of ``PhoneBookEntry`` for each entity in the PhoneBook with a matching Computer Center ID (CCID).
        """
        return self.search("(&{})".format(QUERIES['ccid']).format(value), raw=raw)

    def search_by_uid(self, value: str, raw: bool = False) -> Union[List[PhoneBookEntry], List[Any]]:
        """
        Returns a list of ``PhoneBookEntry`` for each entity in the PhoneBook with a matching UID.
        """
        return self.search("(&{})".format(QUERIES['uid']).format(value), raw=raw)

    def search_by_gid(self, value: str, raw: bool = False) -> Union[List[PhoneBookEntry], List[Any]]:
        """
        Returns a list of ``PhoneBookEntry`` for each entity in the PhoneBook with a matching GID.
        """
        return self.search("(&{})".format(QUERIES['gid']).format(value), raw=raw)

    def search_by_home_directory(self, value: str, raw: bool = False) -> Union[List[PhoneBookEntry], List[Any]]:
        """
        Returns a list of ``PhoneBookEntry`` for each entity in the PhoneBook with a matching home directory.
        """
        return self.search("(&{})".format(QUERIES['home_directory']).format(value), raw=raw)

    @lru_cache(maxsize=None)
    def get_group_info(self, gid: int) -> List[Any]:
        """
        Returns the name of a UNIX group given its GID. This function is memoized to reduce calls to LDAP.
        """
        results = self._raw_search("(gidNumber={})".format(gid), search_base="ou=unix,ou=workgroups,dc=cern,dc=ch", attributes=["name"])
        if results:
            return results[0][1]
        return None
