"""
Represents one PhoneBook entry from the results of an LDAP query.
"""
from typing import Any, Dict, List, Mapping, Optional, Union

import re
import json
import traceback
import collections
from textwrap import dedent
from datetime import datetime


#: Fields that can be passed to the ``--terse`` flag, mapped to their corresponding attributes.
_TERSE_ALIASES = {
    "surname": "last_name",
    "firstname": "first_name",
    "login": "login_name",
    "department": "department",
    "group": "_extensionAttribute13",
    "cerngroup": "group",
    "cernsection": "section",
    "displayname": "_displayName",
    "phone": "landline_phone",
    "otherphone": "other_phone",
    "mobile": "mobile_phone",
    "fax": "fax",
    "office": "office",
    "pobox": "post_box",
    "email": "email",
    "organization": "organization",
    "ccid": "ccid",
    "homedir": "home_directory",
    "last": "_extensionAttribute14",
    "uid": "uid",
    "gid": "gid",
    "uac": "_userAccountControl",
    "type": "account_type",
    "emailnick": "_amilNickname",
    "company": "company",
    "shell": "login_shell",
    "secid": "security_id",
}


class PhoneBookEntry:
    """
    Represents one PhoneBook entry from the results of an LDAP query.

    The data is stored in three ways:

         * **Snake case**, meaningfully-named attributes which contains the most relevant subset of
           information retrieved from the query. These fields are always lists of strings. Lists of
           references are parsed and stripped of their LDAP syntax for better readability, and
           fields known to be dates are parsed into ``datetime`` objects.

           Example: ``self.login_name``, ``self.office``.

           Full list:
                - ``cern_id``
                - ``login_name``
                - ``full_name``
                - ``first_name``
                - ``last_name``
                - ``email``
                - ``user_principal_name``: usually an email address
                - ``external_email``
                - ``office``
                - ``post_box``
                - ``preferred_language``
                - ``landline_phone``
                - ``fax``
                - ``mobile_phone``
                - ``other_phone``
                - ``website``
                - ``info``
                - ``employee_id``
                - ``employee_type``
                - ``computer_center_id``
                - ``ccid``
                - ``gid``
                - ``uid``
                - ``user_account_control``
                - ``cern_low_uid``
                - ``security_id``
                - ``department``
                - ``group``
                - ``section``
                - ``building``
                - ``institute_name``
                - ``institute_abbreviation``
                - ``company``
                - ``devices``
                - ``manager``
                - ``egroups``
                - ``login_shell``
                - ``home_directory``
                - ``last_changed``
                - ``last_login``: it's not really sure this is really the last login date, beware
                - ``created``
                - ``account_status``: active, disabled, locked, ...
                - ``account_type``: Primary or Secondary
                - ``entry_status``: User (a person) or Contact (a service)
                - ``domain``: searched LDAP domain: one of the keys of the ``DOMAINS`` dictionary.

         * **Camel case**, LDAP-named attributes which contain all the data retrieved from the query,
           regardless of redundancy, but parsed for homogeneity. The names are prefixed with an underscore
           (like ``_cn``) to prevent collisions with the snake case fields. These fields are always lists
           of strings too. Lists of references are not parsed and are stored as a list of strings. Dates aren't
           converted either and stored as strings as they came from LDAP.

           Example: ``self._sAMAccountName``, ``self._cn``

           A full list of these fields is unavailable: the fields returned by the LDAP server may vary.
           However, you can convert this object into a dictionary using ``vars(my_object)`` and then search into
           its keys.

         * **Raw data** stored in a single string. It contains the raw results of the LDAP query as-is. Good
           for debugging and for the cases where the fields parsing fails for any reason.

           It's only one field: ``self.raw_data``

    Other fields:

        - ``self.parsing_succeeded``: ``True`` if the parsing was successful, ``False`` otherwise.
            This field is never a list.

        - ``self.exception_raised``: Contains a string representation exception which made the parsing fail,
            or ``None``.
    """

    #: The most relevant data extracted from the LDAP query, with a more readable name.
    KNOWN_FIELDS = {
        'cern_id': "cn",
        'login_name': 'sAMAccountName',
        'full_name': 'displayName',
        'first_name': 'givenName',
        'last_name': 'sn',
        'email': 'mail',
        'user_principal_name': 'userPrincipalName',
        'external_email': 'cernExternalMail',
        'office': 'physicalDeliveryOfficeName',
        'post_box': 'postOfficeBox',
        'preferred_language': 'preferredLanguage',
        'landline_phone': 'telephoneNumber',
        'fax': 'facsimileTelephoneNumber',
        'mobile_phone': 'mobile',
        'other_phone': 'otherTelephone',
        'website': 'wWWHomePage',
        'info': 'info',

        'employee_id': 'employeeID',
        'employee_type': 'employeeType',
        'computer_center_id': 'employeeID',
        'ccid': 'employeeID',
        'gid': 'gidNumber',
        'gid_name': 'gidName',
        'uid': 'uidNumber',
        'user_account_countrol': 'userAccountControl',
        'cern_low_uid': 'cernLowUid',
        'security_id': 'altSecurityIdentities',

        'department': 'division',
        'group': 'cernGroup',
        'section': 'cernSection',
        'building': 'physicalDeliveryOfficeName',
        'institute_name': 'cernInstituteName',
        'institute_abbreviation': 'cernInstituteAbbreviation',
        'company': 'company',

        'devices': "directReports",
        'manager': "manager",
        'egroups': "memberOf",

        'login_shell': 'loginShell',
        'home_directory': 'unixHomeDirectory',
        'last_login': 'extensionAttribute14',

        'last_changed': 'whenChanged',
        'created': 'whenCreated',
        'account_status': 'userAccountControl',
        'account_type': 'employeeType',
        'entry_type': 'objectClass',
    }
    #: Fields known to contain list of references:
    #: these field's content will be parsed to strip LDAP syntax
    #: and return readable, CLI-friendly values.
    KNOWN_REFERECE_FIELDS = [
        'devices',
        'manager',
        'egroups',
    ]
    #: Fields known to contain a date:
    #: These values are parsed according to the specified format.
    KNOWN_DATE_FIELDS = {
        'last_changed': '%Y%m%d%H%M%S.0Z',
        'created': '%Y%m%d%H%M%S.0Z',
        'last_login': "%A %d %B %Y at %I:%M %p",
    }

    def __init__(self, raw_data, domain: Optional[str] = None, verbose: bool = False):
        """
        Parses the data in input and tries to extract all known information from it.
        Stores raw data and unparsed fields too, and records parsing outcomes (success/failure).
        See the documentation of ``PhoneBookEntry`` for a better overview of the parsing strategy.

        :param raw_data: Entry returned from a LDAP query, as-is.
        """
        # Stores raw data as-is
        self.raw_data = str(raw_data)
        self.parsing_succeeded = False
        self.exception_raised = []

        # Store the domain this entry comes from (if given)
        if domain:
            self.domain = [domain]

        try:
            # Pre-processes the data to transform it into a dictionary of lists for easier data extraction
            data = self._parse_ldap_data(raw_data[1])

            # Process known fields programmatically
            for field_name, field_tag in self.KNOWN_FIELDS.items():
                field_value = data.get(field_tag, [])

                # Extra processing step for lists of references
                if field_name in self.KNOWN_REFERECE_FIELDS:
                    field_value = self._process_list_of_references(field_value)

                # Extra processing step for lists of dates
                if field_name in self.KNOWN_DATE_FIELDS.keys():
                    field_value = self._process_list_of_dates(field_value, self.KNOWN_DATE_FIELDS[field_name])

                # Set the attribute
                setattr(self, field_name, field_value)

            # Process remaining fields as-is (won't be printed, but will be available)
            for raw_field_tag, raw_field_value in data.items():
                setattr(self, "_{}".format(raw_field_tag), raw_field_value)

            # Set the last login date to "-" to avoid problems with rendering down the line
            # NOTE: I'm not really sure this is the last login date, so I won't put it in the named fields
            if not self._extensionAttribute14:
                self.extensionAttribute14 = "-"

            # Translate a few attributes
            self.account_status = [self._decode_user_account_status(self._userAccountControl)]

            # Clean up other phones
            self._otherTelephone = self._cleanup_other_phone_values()

            # Create the login string for the --all version
            try:
                self._unix_account_description = [self._create_computer_account_string()]
            except Exception:
                if verbose:
                    print("-" * 80)
                    traceback.print_exc()
                    print("-" * 80)
                self._unix_account_description = []

            self.parsing_succeeded = True

        except Exception as e:
            if verbose:
                print("-" * 80)
                traceback.print_exc()
                print("-" * 80)
            self.parsing_succeeded = False
            self.exception_raised = str(e)

    def __getattr__(self, item):
        """
        Allows to query for LDAP values without having to make sure they are defined every time.

        Returns ``None`` if the requested attribute is not defined.

        :param item: the key to search
        :return: the value of the attribute if found, ``None`` otherwise
        """
        if item in self.KNOWN_FIELDS:
            return self.KNOWN_FIELDS[item]

    def __repr__(self) -> str:
        """
        Do not use for pretty printing. Use __str__ instead.

        :return: Unique, concise and fail_safe data representation to tell the various objects apart.
        """
        try:
            return "<PhoneBookEntry object for {} at {}>".format(self.login_name, id(self))
        except AttributeError:
            return "<PhoneBookEntry object for <no-login-name> at {}>".format(id(self))

    def __str__(self) -> str:
        return self._render_one_line()

    def _render(self, all_fields: bool = False, terse: Optional[List[str]] = None, json: Optional[List[str]] = None):
        """
        Renders an entry in different ways according to the parameters given. Can render:

        1. As a single line, with some spaced out values, so that multiple entries resemble a table
        2. As a full record, with detailed info spaced in a way that you can see most of the data at a glance
        3. As a computer-friendly record, with the fields specified in the ``terse`` parameter separated by a semicolon.
        4. As JSON

        This function exists only for the backward-compatible CLI. Do not use in your code.
        PhoneBook can output JSON instead, which contains a lot more information on the entry as well.

        :param all_fields: True to get a full record card of the entry, ``False`` to get a one-liner. Default ``False``.
        :param terse: if given, it renders the entry as a list of information from the specified fields, in order,
            separated by a semicolon. ``terse`` accepts only a limited amount of values, which you can see in
            ``TERSE_ALIASES``. Please consider using the json facilities instead, or directly the Python API,
            if you're building a script based on the PhoneBook.
        :return: Returns a string with the desired representation of the entry.
        """
        if all_fields:
            return self._render_complete()
        elif terse:
            return self._render_terse(terse)
        elif json:
            return self._render_json(json)
        else:
            return self._render_one_line()

    def _render_one_line(self) -> str:
        """
        Renders the entry as a single line, with some spaced out values, so that multiple entries resemble a table.

        This function exists only for the backward-compatible CLI. Do not use in your code.
        PhoneBook can output JSON instead, which contains a lot more information on the entry as well.

        :return: A printable, one-line data representation of this ``PhoneBookEntry``.
        """
        if not self.parsing_succeeded:
            return f" - The parsing failed for this user! Exception: {self.exception_raised}"

        last_names = self._render_field(self.last_name, first=True)
        first_names = self._render_field(self.first_name, first=True)
        landline_phone = self._render_field(self._telephoneNumber, short_phone_number=True)
        other_phone = self._render_field(self._otherTelephone, short_phone_number=True, first=True)
        mobile = self._render_field(self._mobile, short_phone_number=True)
        department = self._render_field(self.department, first=True)
        group = self._render_field(self.group, first=True)
        section = self._render_field(self.section, first=True)
        office = self._render_field(self.office, first=True)

        # Syntax note: ' <24.24s' means "pad with spaces on the right up to 24 chars if necessary" (' <24')
        # and "limit to 24 chars if the string is longer than that" ('.24s')
        return f"{last_names: <24.24s} " \
               f"{first_names: <15.15s} " \
               f"{landline_phone: <5.5s} " \
               f"{other_phone: <5.5s} " \
               f"{mobile: <6.6s} " \
               f"{department: <3.3s} " \
               f"{group: <3.3s} " \
               f"{section: <5.5s} " \
               f"{office: >9.11s}"

    def _render_terse(self, fields: List[str]):
        """
        Renders the entry as a computer-friendly record, with the fields specified in the ``terse``
        parameter separated by a semicolon.

        This function exists only for the backward-compatible CLI. Do not use in your code.
        PhoneBook can output JSON instead, which contains a lot more information on the entry as well.

        :param fields: the fields to render, from the keys of ``TERSE_ALIASES``
        :return: A custom printable representation of this ``PhoneBookEntry``
        """
        actual_fields = [_TERSE_ALIASES[field] for field in fields if field in _TERSE_ALIASES.keys()]

        field_values = []
        for field in actual_fields:
            values = getattr(self, field)
            value = " ".join(values or ["-"])
            field_values.append(value)

        return ";".join(field_values) + ";"

    def _render_json(self, fields: List[str]):
        """
        Renders the entry as JSON, with the fields specified in the ``fields``
        parameter separated by a semicolon.

        :param fields: the fields to render, from the keys of ``TERSE_ALIASES``
        :return: A JSON representation of this ``PhoneBookEntry``
        """
        result = {}
        for f in fields:
            result[f] = getattr(self, _TERSE_ALIASES[f])

        return json.dumps(result)

    def _render_complete(self) -> str:
        """
        Renders the entry as a full record, with detailed info spaced in a way that you can see most of the data
        at a glance.

        This function exists only for the backward-compatible CLI. Do not use in your code.
        PhoneBook can output JSON instead, which contains a lot more information on the entry as well.

        :return: A printable, complete data representation of a ``PhoneBookEntry``.
        """
        # Render optional email fields
        other_email = ['\nOther E-mail:       '+email for email in self.email[1:] if len(self.email) > 1]
        other_email = "".join(other_email)
        external_email = '\nExternal E-mail:    '+self._render_field(self.external_email) if self.external_email else ""

        # Render organization names
        organizations = []
        if self.institute_name:
            organizations += self.institute_name
        if self.company:
            organizations += self.institute_name
        organizations = list(dict.fromkeys(organizations)) or ["-"]
        organization = "\n" + "\n".join([f'Organization:       {o}' for o in organizations])

        # Render accounts
        accounts = '\n'.join(self._unix_account_description)

        # Render common sections of the record, and define those that might be empty
        emails_part = f"E-mail:             {self._render_field([self.email[0]])}" + other_email + external_email
        phones_part = dedent(f"""\
            Telephone:          {self._render_field(self._telephoneNumber): <41.41s} (internal: {self._render_field(
                self._telephoneNumber, short_phone_number=True): >6.6s})""")

        if self._otherTelephone:
            phones_part += dedent(f"""
            Other telephone:    {self._render_field(self._otherTelephone, first=True): <41.41s} (internal: {self._render_field(
                self._otherTelephone, short_phone_number=True, first=True): >6.6s})""")

        phones_part += dedent(f"""
            Mobile:             {self._render_field(self.mobile_phone): <41.41s} (internal: {self._render_field(
                self.mobile_phone, short_phone_number=True): >6.6s})
            Facsimile:          {self._render_field(self.fax): <41.41s} (internal: {self._render_field(
                self.fax, short_phone_number=True): >6.6s})""")
        organic_unit_part = ""
        office_part = f"""Bld. Floor-Room:    {self._render_field(self.office)}{organization}"""
        info_part = ""
        computing_id_part = ""
        logins_part = ""

        # Render sections typical of services: service name, webpage, etc...
        if "contact" in self.entry_type:
            names_part = dedent(f"""\
                Service:            {self._render_field(self.full_name)}
                Service:            {self._render_field(self.company)} """)
            info_part = dedent(f"""\
                Information:        {self._render_field(self.info)}
                WWW:                {self._render_field(self.website)} """)

        # Render sections typical of users: name, organic unit, computing IDs, etc...
        else:
            names_part = dedent(f"""\
                Surname:            {self._render_field(self.last_name, first=True)}
                Firstname:          {self._render_field(self.first_name, first=True)}
                Display Name:       {self._render_field(self.full_name, first=True)}""")
            organic_unit_part = dedent(f"""\
                Department:         {self._render_field(self.department, first=True)}
                Group:              {self._render_field(self.group, first=True)}
                Section:            {self._render_field(self.section, first=True)}
                POBox:              {self._render_field(self.post_box): <41.41s} (internal mail)""")
            computing_id_part = dedent(f"""\
                Computer Center ID: {self._render_field(self.ccid)}
                Low user ID:        {self._render_field(self.cern_low_uid)}""")

            # Only users have UNIX login information, so render only if this is one
            if "users" in self.domain:
                logins_part = dedent(f"""\

                Computer account(s):
                Login    Grp    St Uid    Gid  Last login     Shell      Home directory\n\n""") + accounts

        return "#------------------------------------------------------------------------------\n" + \
               "\n".join([line for line in [
                   names_part,
                   emails_part,
                   phones_part,
                   organic_unit_part,
                   office_part,
                   info_part,
                   computing_id_part,
                   logins_part] if line != ""])

    def _render_field(self, values: List[str], empty_value: str = "-", short_phone_number: bool = False, first: bool = False):
        """
        Transforms a list of values into a single string, space separated. If the list was empty,
        returns a dash.

        :param values: a list of strings
        :param empty_value: what to return if the list is empty
        :param short_phone_number: whether the fields are phone numbers that should be shortened
        :param first: don't concatenate values, just use the first one on the list
        :return: a single string with the rendered values
        """
        if not values:
            return empty_value
        if short_phone_number:
            values = [self.long_phone_number_to_short(p) for p in values]
        if first:
            return values[0]
        return " ".join(values)

    def _parse_ldap_data(self, data: Mapping[str, Any]) -> Mapping[str, List[str]]:
        """
        Transforms the result of a LDAP query into a dictionary with all values as lists of strings.
        In fact, theoretically any field from LDAP could be a list.

        :param data: the raw data from the LDAP query.
        :return: a dictionary of list of strings containing all the data found in the LDAP response.
        """
        parsed_data = {}
        for key, raw_values in data.items():

            # If a value is None, it will become an empty list.
            # Not sure any value ever comes as None, but better safe than sorry
            if raw_values is None:
                parsed_data[key] = []
                continue

            values = []
            # If the raw_value is an Iterable but not a string:
            if isinstance(raw_values, collections.Iterable) and not isinstance(raw_values, str):
                # Append values one by one after decoding them
                for value in raw_values:
                    values.append(self._fail_safe_decode(value))
            else:
                # Decode the entire value and append it
                values.append(self._fail_safe_decode(raw_values))
            parsed_data[key] = values

        return parsed_data

    @staticmethod
    def _fail_safe_decode(value: Any) -> str:
        """
        Strings mostly come encoded as bytes, but sometime not.
        This decoded ensures the result is a regular string.

        :param value: Potentially encoded string
        :return: Surely decoded string
        """
        try:
            return value.decode()
        except (AttributeError, UnicodeDecodeError):
            return str(value)

    def _create_computer_account_string(self) -> str:
        """
        Create the string seen at the bottom of the ``--all`` view. This function is used only for the
        backward-compatible CLI and is not stable API.

        These lines should be created at parsing time, because later on multiple PhoneBook entries
        might be merged by CCID and finding out the original pairs can be messy.

        :return: The string seen at the bottom of the ``--all`` view in the backward-compatible CLI,
            with the info on the computing accounts of the person.
        """
        login_name = self.login_name[0] if len(self.login_name) > 0 else "-"
        gid_name = self.gid_name[0] if len(self.gid_name) > 0 else "--"
        tipe = self._employeeType[0] if len(self._employeeType) > 0 else "?"
        status = self.account_status[0] if len(self.account_status) > 0 else "?"
        uid = self.uid[0] if len(self.uid) > 0 else "??"
        gid = self.gid[0] if len(self.gid) > 0 else "??"
        date = self.last_login[0].strftime("%d/%m/%y %H:%M") if len(self.last_login) > 0 else "--/--/-- --:--"
        shell = self.login_shell[0] if len(self.login_shell) > 0 else "-"
        home = self.home_directory[0] if len(self.home_directory) > 0 else "/"

        desc = f"{login_name: <8.8s} " \
               f"{gid_name: <6.6s} " \
               f"{tipe[0].upper()}{status[0].upper()} " \
               f"{uid: <6.6s} " \
               f"{gid: <4.4s} " \
               f"{date: <14.14s} " \
               f"{shell: <10.10s} " \
               f"{home}"
        return desc

    def _cleanup_other_phone_values(self) -> List[str]:
        """
        ``self._otherTelephone`` is quite polluted in many cases: remove duplicates and normalize to short form.

        :return: a list with the cleaned up list of other phone numbers
        """
        telephone_numbers = self._telephoneNumber or []
        other_phones = self._otherTelephone or []

        if other_phones and telephone_numbers:
            telephone_numbers = [self.short_phone_number_to_long(p) for p in telephone_numbers]
            other_phones = [self.short_phone_number_to_long(p) for p in other_phones
                            if self.short_phone_number_to_long(p) not in telephone_numbers]
            return list(dict.fromkeys(other_phones))
        return [self.short_phone_number_to_long(p) for p in other_phones]

    @staticmethod
    def _process_list_of_references(data: List[str]) -> Optional[List[str]]:
        """
        Converts a list of strings from::

            [CN=devicename1,OU=CERN Linux Computers,DC=cern,DC=ch,
             CN=devicename2,OU=CERN Linux Computers,DC=cern,DC=ch]

        to::

            [devicename1, devicename2]

        :param data: the list of references to process
        :return: a list of readable strings, or ``None`` is ``data`` was ``None``
        """
        if data is None:
            return None
        try:
            return [entry.split(",")[0].split("=")[1] for entry in data]
        except KeyError:
            return data

    @staticmethod
    def _process_list_of_dates(dates: Optional[List[str]], format_string: str) \
            -> Union[List[datetime], List[str], None]:
        """
        Converts a list of dates from strings into ``datetime`` objects, applying the
        ``format_string`` format to all of them. Fails silently, returning the
        unconverted values, if ``ValueError`` is raised, i.e. if the date string doesn't
        match the date format.

        :param dates: the list of string dates to process
        :param format_string: the format to apply to the date
        :return: a list of datetime objects, or ``None`` is ``data`` was ``None``
        """
        if dates is None:
            return None
        try:
            return [datetime.strptime(entry, format_string) for entry in dates]
        except ValueError:
            return dates

    @staticmethod
    def _decode_user_account_status(status_code: List[str]) -> str:
        """
        Translate the ``self._userAccountControl`` field according to the instructions found at
        http://support.microsoft.com/kb/305144

        :param status_code: The content of the field ``self._userAccountControl``, straight from LDAP
        :return: ``active``, ``disabled``, ``password_expired``, ``locked`` or ``unknown (reason)``
        """
        if not status_code:
            return f"unknown (no value)"
        else:
            status = status_code[0]
            if status == '512':
                return "active"
            elif status == "2" or status == "514":  # 514 is nonstandard but it's full of these codes...
                return "disabled"
            elif status == "8388608":
                return "password_expired"
            elif status == "16":
                return "locked"
            else:
                return f"unknown ({status})"

    @staticmethod
    def long_phone_number_to_short(value: str) -> str:
        """
        Translates long phone numbers into their short version.

        Rough translation of a Perl subroutine ``trph`` found in the original
        ``phonebook`` code (https://gitlab.cern.ch/linuxsupport/rpms/phonebook/-/blob/master/src/phonebook.pl#L28)

        Please fix bugs in case you find any.

        :param value: the phone number.
        :return: the shortened phone number.
        """
        if not value:
            return "-"
        match = re.search(r"^(\+|00)412276(\d\d\d\d\d)$", value)
        if match:
            return match.group(2)
        match = re.search(r"^(\+|00)4175411(\d\d\d\d)$", value)
        if match:
            return "16{}".format(match.group(2))
        match = re.search(r"^(\+|00)412276(7XXXX$)", value)
        if match:
            return match.group(2)
        return value

    @staticmethod
    def short_phone_number_to_long(value: str) -> str:
        """
        Translates short phone numbers into their complete version.

        Rough translation of a Perl subroutine ``trph`` found in the original
        ``phonebook`` code (https://gitlab.cern.ch/linuxsupport/rpms/phonebook/-/blob/master/src/phonebook.pl#L28)

        Please fix bugs in case you find any.

        :param value: the phone number.
        :return: the complete phone number.
        """
        if not value:
            return "-"
        match = re.search(r"^16(\d\d\d\d)$", value)
        if match:
            return "+4175411{}".format(match.group(1))
        match = re.search(r"^(\d\d\d\d\d)$", value)
        if match:
            return "+412276{}".format(match.group(1))
        # Not needed? already matched by the previous check
        # match = re.search(r"^(7\d\d\d\d)$", value)
        # if match:
        #     return "+412276{}".format(match.group(1))
        return value

    def to_dict(self, include_raw: bool = False) -> Dict[str, List[str]]:
        """
        Returns a cleaned up version of the values as a dict, plus the original data
        in the ``raw_data`` field if ``raw==True``.

        :param include_raw: include the ``raw_data`` field into the resulting JSON if True.
        :return: a JSON string containing all the object's fields.
        """
        # Either contains info about the failure, or contains all the data
        if not self.parsing_succeeded:
            dictionary = {"parsing_succeeded": False,
                          "exception_raised": self.exception_raised}
        else:
            dictionary = vars(self)

        # Either make sure to include the raw_data field
        if include_raw:
            dictionary["raw_data"] = self.raw_data
        # Or make sure it's not present
        else:
            dictionary.pop("raw_data", None)

        return dictionary

    @staticmethod
    def json_serializer(data):
        """
        JSON serializer used to dump the entry. It only converts dates into strings.

        :param data: value to translate, if necessary, from datetime to string
        :return: a string with the value of the datetime object.
        """
        if isinstance(data, datetime):
            return data.__str__()

    def to_json(self, include_raw: bool = False, *args, **kwargs) -> str:
        """
        Returns a cleaned up version of the values as JSON, plus the original data
        in the ``raw_data`` field if ``raw==True``.

        Extra parameters are passed to ``json.dumps`` directly.

        Dates are automatically serialized as strings with ``self.json_serializer``.

        :param include_raw: include the ``raw_data`` field into the resulting JSON if True.
        :return: a JSON string containing all the object's fields.
        """
        to_dump = self.to_dict(include_raw=include_raw)
        return json.dumps(to_dump, default=self.json_serializer, *args, **kwargs)
