import json
import os.path
from typing import Any, List, Union

import ldap3, pytest
import jq
from pytest_console_scripts import ScriptRunner, RunResult
from pytest_mock import MockerFixture

from pyphonebook import PhoneBook, PhoneBookEntry



@pytest.fixture
def mock_ldap(mocker: MockerFixture) -> None:
    print('Mocking LDAP server')
    def fake_server(self, server):
        fake_connection = ldap3.Connection(ldap3.Server('fake_server'), client_strategy=ldap3.MOCK_SYNC)
        fake_data_path = os.path.join(os.path.dirname(__file__), 'fake_data.json')
        fake_connection.strategy.entries_from_json(fake_data_path)
        fake_connection.bind()

        return fake_connection

    mocker.patch('pyphonebook.PhoneBook._getLDAP', new=fake_server)


@pytest.fixture
def phonebook(mock_ldap: pytest.fixture) -> PhoneBook:
    pb = PhoneBook()
    yield pb


@pytest.fixture
def mock_users(phonebook: pytest.fixture) -> Union[List[PhoneBookEntry], List[Any]]:
    yield phonebook.search("(cn=*)", raw=True)

# Convenience function to make it simpler to specify command lines
def command(self: ScriptRunner, cmd: str = '') -> RunResult:
    cmd = 'pyphonebook {}'.format(cmd).split()
    return self.run(*cmd)
ScriptRunner.command = command

# Reduce the boilerplate in checking script output
def check(self: ScriptRunner, command: str = '', stdout: List[str] = [], stderr: List[str] = [], returncode: int = 0) -> RunResult:
    ret = self.command(command)
    assert ret.returncode == returncode

    if stderr:
        assert ret.stderr == '\n'.join(stderr)+'\n'
    else:
        assert ret.stderr == ''

    if stdout:
        assert ret.stdout == '\n'.join(stdout)+'\n'
    else:
        assert ret.stdout == ''

    return ret
ScriptRunner.check = check

def check_json(self: ScriptRunner, command: str = '', stdout_json: str = '', stderr: List[str] = [], returncode: int = 0) -> RunResult:
    ret = self.command(command)
    assert ret.returncode == returncode

    if stderr:
        assert ret.stderr == '\n'.join(stderr)+'\n'
    else:
        assert ret.stderr == ''

    # We load the string to test with json, but we compare the output using
    # jq. This way we check that the json really is valid, just in case there's
    # some weird bug.
    try:
        data = json.loads(stdout_json)
    except json.JSONDecodeError as e:
        pytest.fail(f'Test json string is invalid: {e}')
    keys = sorted([k for keys in data for k in keys])

    # Basic test to see if the json is valid or not
    try:
        json.loads(ret.stdout)
    except json.JSONDecodeError as e:
        pytest.fail(f'Command output is not valid json: {e}')

    # Check that we got an array
    assert jq.compile('type').input(text=ret.stdout).first() == "array"

    # Check that we have all the same keys
    assert sorted(jq.compile('.[] | keys[]').input(text=ret.stdout).all()) == keys

    # Check that all keys have the expected values
    for k in keys:
        assert sorted(jq.compile(f'.[].{k}').input(text=ret.stdout).all()) == sorted([x[k] for x in data])

    return ret
ScriptRunner.check_json = check_json

@pytest.fixture
def pyphonebook(mock_ldap: pytest.fixture, script_runner: ScriptRunner) -> ScriptRunner:
    yield script_runner
