"""
High-level tests for the PhoneBookEntry class.

"""
from pyphonebook import PhoneBookEntry


def test_valid(mock_users):
    for user in mock_users:
        entry = PhoneBookEntry(user)
        assert entry.parsing_succeeded
        assert "exception_raised" in vars(entry).keys()
        assert entry.exception_raised == []
        assert "raw_data" in vars(entry).keys()
        assert entry.to_json()
        assert str(entry)
        assert entry.login_name[0] == user[1]["cn"][0]


def test_unexpected_but_valid():
    entry = PhoneBookEntry(("broken", {"unexpected": "value"}))
    assert entry.parsing_succeeded
    assert "exception_raised" in vars(entry).keys()
    assert entry.exception_raised == []
    assert "raw_data" in vars(entry).keys()
    assert entry.to_json()
    assert str(entry)
    assert entry.login_name == []


def test_broken():
    entry = PhoneBookEntry("I should not be a string!")
    assert not entry.parsing_succeeded
    assert "exception_raised" in vars(entry).keys()
    assert entry.exception_raised
    assert "raw_data" in vars(entry).keys()
    assert entry.to_json()
    assert str(entry)
    assert "login_name" not in vars(entry).keys()


def test_long_phone_number_to_short_noop():
    assert "41227697554" == PhoneBookEntry.long_phone_number_to_short("41227697554")

def test_long_phone_number_to_short_landline():
    assert "97554" == PhoneBookEntry.long_phone_number_to_short("+41227697554")
    assert "74258" == PhoneBookEntry.long_phone_number_to_short("0041227674258")

def test_long_phone_number_to_short_mobile():
    assert "165338" == PhoneBookEntry.long_phone_number_to_short("+41754115338")
    assert "162738" == PhoneBookEntry.long_phone_number_to_short("0041754112738")

def test_long_phone_number_to_short_fax():
    assert "69018" == PhoneBookEntry.long_phone_number_to_short("+41227669018")
    assert "69518" == PhoneBookEntry.long_phone_number_to_short("0041227669518")


def test_short_phone_number_to_long_noop():
    assert "41227697554" == PhoneBookEntry.short_phone_number_to_long("41227697554")

def test_short_phone_number_to_long_landline():
    assert "+41227697553" == PhoneBookEntry.short_phone_number_to_long("97553")
    assert "+41227674158" == PhoneBookEntry.short_phone_number_to_long("74158")

def test_short_phone_number_to_long_mobile():
    assert "+41754115328" == PhoneBookEntry.short_phone_number_to_long("165328")
    assert "+41754111338" == PhoneBookEntry.short_phone_number_to_long("161338")

def test_short_phone_number_to_long_fax():
    assert "+41227663018" == PhoneBookEntry.short_phone_number_to_long("63018")
    assert "+41227669618" == PhoneBookEntry.short_phone_number_to_long("69618")


def test_render_one_line(mock_users):
    EXPECTED = {
        "secondary1": "Lastname1                User            71234 74257 161234 IT  CM  LCS    99 1-010",
        "dcognome1" : "Cognome1                 Dh1             61111 62901 162111 IT  -   -      91 2-027",
        "user1"     : "Lastname1                User            72222 71111 161234 IT  CM  LCS    99 1-010",
        "user3"     : "van Lastname3            User            -     -     -      IT  CM  COG    99 1-005",
        "user4"     : "Vastname4                User            -     -     161135 BE  CSS SET   799 2-034",
        "user5"     : "Lastname5                User            -     -     -      Oth -   -             -",
        "user6"     : "Lastname6                User            71153 -     164658 IT  CM  -      99 R-008",
        "esiuser2"  : "User1                    esi             -     -     -      Oth -   -             -",
    }
    for user in mock_users:
        entry = PhoneBookEntry(user)
        assert EXPECTED[entry.login_name[0]] == entry._render_one_line()

def test_render_json(mock_users):
    EXPECTED = {
        "secondary1": '{"firstname": ["User"], "department": ["IT"], "phone": ["+41227671234"], "office": ["99 1-010"]}',
        "dcognome1" : '{"firstname": ["Dh1"], "department": ["IT"], "phone": ["+41227661111"], "office": ["91 2-027"]}',
        "user1"     : '{"firstname": ["User"], "department": ["IT"], "phone": ["+41227672222"], "office": ["99 1-010"]}',
        "user3"     : '{"firstname": ["User"], "department": ["IT"], "phone": [], "office": ["99 1-005"]}',
        "user4"     : '{"firstname": ["User"], "department": ["BE"], "phone": [], "office": ["799 2-034"]}',
        "user5"     : '{"firstname": ["User"], "department": ["Other"], "phone": [], "office": []}',
        "user6"     : '{"firstname": ["User"], "department": ["IT"], "phone": ["+41227671153"], "office": ["99 R-008"]}',
        "esiuser2"  : '{"firstname": ["esi"], "department": ["Other"], "phone": [], "office": []}',
    }
    for user in mock_users:
        entry = PhoneBookEntry(user)
        assert EXPECTED[entry.login_name[0]] == entry._render_json(["firstname", "department", "phone", "office"])

# TODO test rendering (_render, _render_terse, _render_complete)
# TODO test _cleanup_other_phone_values

def test_decode_user_account_status():
    # `active``, ``disabled``, ``password_expired``, ``locked`` or ``unknown (reason)``
    # 512, 2 (514), 8388608, 16, X
    assert "active" == PhoneBookEntry._decode_user_account_status(["512"])
    assert "disabled" == PhoneBookEntry._decode_user_account_status(["2"])
    assert "disabled" == PhoneBookEntry._decode_user_account_status(["514"])
    assert "password_expired" == PhoneBookEntry._decode_user_account_status(["8388608"])
    assert "locked" == PhoneBookEntry._decode_user_account_status(["16"])
    assert "unknown (42)" == PhoneBookEntry._decode_user_account_status(["42"])
